import{r as o,Q as x,se as t,sf as s,a5 as n,sg as v,sh as a,si as c}from"./index-RIg2rYDo.js";function l({children:r}){const{i18n:p}=x(),e=p.dir(),i=o.useMemo(()=>t({...s,direction:e,shadows:{...s.shadows,selected:"inset 0px 0px 0px 3px var(--invoke-colors-invokeBlue-500), inset 0px 0px 0px 4px var(--invoke-colors-invokeBlue-800)",hoverSelected:"inset 0px 0px 0px 3px var(--invoke-colors-invokeBlue-400), inset 0px 0px 0px 4px var(--invoke-colors-invokeBlue-800)",hoverUnselected:"inset 0px 0px 0px 2px var(--invoke-colors-invokeBlue-300), inset 0px 0px 0px 3px var(--invoke-colors-invokeBlue-800)",selectedForCompare:"inset 0px 0px 0px 3px var(--invoke-colors-invokeGreen-300), inset 0px 0px 0px 4px var(--invoke-colors-invokeGreen-800)",hoverSelectedForCompare:"inset 0px 0px 0px 3px var(--invoke-colors-invokeGreen-200), inset 0px 0px 0px 4px var(--invoke-colors-invokeGreen-800)"}}),[e]);return o.useEffect(()=>{document.body.dir=e},[e]),n.jsx(v,{theme:i,toastOptions:a,children:n.jsx(c,{children:r})})}const d=o.memo(l);export{d as default};
